/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Local editing functions -- machine dependent module
 *
 * This module only works on 1 file at a time.
 * The master program is responsible for making sure of that!
 * Other routines are provided to open multiple files, but the
 * caller is responsible for keeping track of streams, etc.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <pwd.h>
#include <sys/param.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <unistd.h>
#include <string.h>

#include "common.h"
#include "and_file.h"
#include "and_scr.h"
#include "ezconfig.h"
#include "xmalloc.h"

extern char *getenv(), *index();

#define MAX_EXEC_ARGS 20

long FIO_GetFileSize(), FIO_GetModifyTime();

FILE *g_file = (FILE *)NULL;

char g_fname[MAXPATHLEN];

void FIO_SetTextMode(int on_off)
{
  /* we always read in binary mode in Unix.  This is useful for machines
     such as PCs for CR-LF <-> LF translation.
  */
}

char *FIO_GenUniqueName(void)
{
  static char unique_filename[MAXPATHLEN];
  struct timeval tp;
  struct timezone tz;
  gettimeofday(&tp, &tz);
  sprintf(unique_filename, "ezm%04d_%08x", (int)getuid(), (unsigned int) tp.tv_sec);
  return(&unique_filename[0]);
}

int FIO_GenTmpName(char *fbuf)
{
    char path_buf[MAXPATHLEN];
    const char *tmpdir = config_getstring("TempDir",NULL);
    if (tmpdir) {
	if (FIO_ResolveName(tmpdir, path_buf) != 0)
	    strcpy(path_buf, tmpdir);
	if (path_buf[strlen(path_buf)-1] != '/') strcat(path_buf, "/");
    } else {
	strcpy(path_buf, "/tmp/");
    }
    sprintf(fbuf, "%s%s", path_buf, FIO_GenUniqueName());
    return(0);
}

int FIO_ExpandName(char *o_fn, char *e_fn,int try_home)
{
  char fbuf[MAXPATHLEN];
  
  if (try_home) {
      strcpy(fbuf, "~/");
      strcat(fbuf, o_fn);
      o_fn = &fbuf[0];
  }
  return(FIO_ResolveName(o_fn, e_fn));
}

int FIO_ResolveName(const char *o_fn,char *e_fn)
{
  if (!o_fn || !*o_fn) return(-1);

  if (*o_fn == '/') strcpy(e_fn, o_fn);
  else {
    char fnbuf[MAXPATHLEN];
    char *fn = &fnbuf[0];
    strcpy(fn, o_fn);
    *e_fn = '\0';
    if (*fn == '~') {
      char *home;
      struct passwd *pwd_s;
      fn++;
      if (*fn == '/') {
	if ((home = getenv("HOME")) == 0) return(-1);
	fn++;
      } else {
	char *uname = index(fn, '/');
	if (uname == NIL) return(-1);
	*uname = '\0';
	if ((pwd_s = getpwnam(fn)) == 0) return(-1);
	home = pwd_s->pw_dir;
	fn = uname + 1;
      }
      strcpy(e_fn, home);
    } else {
      if (getcwd(e_fn, MAXPATHLEN) == 0) return(-1);
      if (strncmp(fn, "./", 2) == 0) fn += 2;
    }
#ifdef RUBBISH
    if (strlen(fn) <= 0) return(-1);
    strcat(e_fn, "/");
    strcat(e_fn, fn);
#else
    strcat(e_fn, "/");
    if (strlen(fn) > 0) strcat(e_fn, fn);
#endif
  }
  return(0);
}

int FIO_CreateFile(char *fname)
{
  if (FIO_ResolveName(fname, g_fname)) return(-1);

  g_file = fopen(g_fname, "w");

  return(g_file == NULL ? -1 : 0);
}

int FIO_WriteToFile(char *buf,int   len)
{
  if ((len = fwrite(buf, 1, len, g_file)) == 0) return(-1);

  return(len);
}

int FIO_ReadFromFile(char *buf,int  size)
{
  int rc;

  rc = fread(buf, 1, size, g_file);
  return(rc);
}

int FIO_OpenFile(char *fname)
{
  if (FIO_ResolveName(fname, g_fname)) return(-1);

  g_file = fopen(g_fname, "r");

  return(g_file == NULL ? -1 : 0);
}

int FIO_CloseFile(void)
{
  int rc;
  rc = fclose(g_file);
  g_file = NULL;
  return(rc);
}

void FIO_EraseFile(char *fname)
{
  unlink(fname);
}

void FIO_SetupArgs(int max_args,char **argbuf,char *pname,char *fname)
{
  /*
    Set up an arguemnt block for exec.  Note that pname gets changed,
    sp pass a copy to it, not the original.
  */
  int i = 0;
  char *ptr;

  max_args -= 2;

  while(max_args > 0 && pname && *pname) {
    ptr = index(pname, ' ');
    if (ptr) *ptr++ = '\0';
    argbuf[i++] = pname;
    pname = ptr;
  }
  if (fname) argbuf[i++] = fname;
  argbuf[i] = (char *)0;
}

int FIO_EditFile(char *fname,int  *changed)
{
  int pid;
  long mtime;
  int status;
  const char *my_favorite_editor = config_getstring("ExternalEditor",NULL);
  char *exec_args[MAX_EXEC_ARGS];
  char editor_name[STR255LEN];

  if (*my_favorite_editor == 0) return(-1);

  strcpy(editor_name, my_favorite_editor);

  FIO_SetupArgs(MAX_EXEC_ARGS, exec_args, editor_name, fname);

  mtime = FIO_GetModifyTime(fname);

  if ((pid = vfork()) == -1) return(-1);

  if (pid == 0) {
    SCR_NoRawState();
    SCR_WriteStatusLine("Starting your favorite editor, please wait...");
    execvp(editor_name, exec_args);
    _exit(-1);
  }

  while(pid > 0) {
    int cid = wait(&status);
    if (cid < 0 || cid == pid) break;
  }

  *changed = ((mtime == FIO_GetModifyTime(fname)) ? FALSE : TRUE);

  SCR_RawState();
  return((WEXITSTATUS(status) != 0) ? -1 : 0);
}

long FIO_GetFileSize(char *fname)
{
  struct stat sbuf;

  if (get_stat(fname, &sbuf) == 0)
    if ((sbuf.st_mode & S_IFDIR) == 0) return((long)sbuf.st_size);

  return(-1L);
}

long FIO_GetModifyTime(char *fname)
{
  struct stat sbuf;

  if (get_stat(fname, &sbuf) == 0)
    if ((sbuf.st_mode & S_IFDIR) == 0) return((long)sbuf.st_mtime);

  return(-1L);
}

int get_stat(char *fname,struct stat *sbuf)
{
  char fnbuf[MAXPATHLEN];
  if (FIO_ResolveName(fname, fnbuf) == 0) {
    return(stat(fnbuf, sbuf));
  }
  return(-1);
}

void FIO_CloseStream(FILE *a_stream)
{
  if (a_stream) fclose(a_stream);
}

FILE *FIO_OpenStream(char *fname,int  flags)
{
  char  efn[MAXPATHLEN], *omode;
  FILE *a_file = (FILE *)NULL;

  if (FIO_ResolveName(fname, &efn[0]) == 0) {
    strcpy(fname, &efn[0]);
    if (flags == OPEN_R) omode = "r";
    else if (flags == OPEN_W) omode = "w";
    else if (flags == OPEN_A) omode = "a";
    else omode = "r+";
    a_file = fopen(fname, omode);
    if (a_file == (FILE *)NULL && errno == ENOENT) {
      if ((a_file = fopen(fname, "w"))!=NULL) {
	fclose(a_file);
	a_file = fopen(fname, omode);
      }
    }
  }
  return(a_file);
}

int FIO_ReadStream(FILE *stream,char *buf,unsigned int len)
{
  FILE *save_g_file = g_file;
  g_file = stream;
  len = FIO_ReadFromFile(buf, len);
  g_file = save_g_file;
  return(len);
}

int FIO_WriteStream(FILE *stream,char *buf,unsigned int len)
{
  FILE *save_g_file = g_file;
  g_file = stream;
  len = FIO_WriteToFile(buf, len);
  g_file = save_g_file;
  return(len);
}

int FIO_SeekStream(FILE *stream,long  pos)
{
  int rc = 0;
  long cur_pos = ftell(stream);

  if (pos != cur_pos) {
    if (pos == 0) rewind(stream);
    else if (pos > 0) rc = fseek(stream, pos, 0);
    else rc = -1;
  }
  return(rc);
}

int FIO_AppendFile(char *f1,char *to_f2,long fsize)
{
  int rc = 0;
  FILE *s1 = (FILE *)NULL, *s2 = (FILE *)NULL;

  if ((s1 = fopen(f1, "r")) == (FILE *)NULL ||
      (s2 = fopen(to_f2, "a")) == (FILE *)NULL) rc = -1;
  else {
    char bigbuf[1024];
    while (fsize > 0) {
      int len = imin(fsize, sizeof(bigbuf)-1);
      bigbuf[0] = 0;
      rc = fread(&bigbuf[0], 1, len, s1);
      if (rc <= 0) {
	if (feof(s1)) rc = 0;
	break;
      }
      fsize -= rc;
      rc = fwrite(&bigbuf[0], 1, rc, s2);
      if (rc <= 0) break;
      rc = 0;
    }
  }

  if (s1) fclose(s1);
  if (s2) fclose(s2);

  return(rc);
}

int FIO_CloneFile(char *f1,char *f2,long  fsize)
{
  char ef1[MAXPATHLEN], ef2[MAXPATHLEN];
  FILE *s2;

  if ((FIO_ResolveName(f1, &ef1[0])) ||
      (FIO_ResolveName(f2, &ef2[0])) ||
      ((s2 = fopen(&ef2[0], "w+")) == (FILE *)NULL)) return(-1);

  fclose(s2);

  return(FIO_AppendFile(&ef1[0], &ef2[0], (long)fsize));
}

int FIO_MakeBackup(char *fname)
{
    char *newfname;
    char data[4096];
    FILE *oldf;
    FILE *newf;

    newfname = xmalloc(strlen(fname)+4+1);
    strcpy(newfname, fname);
    strcat(newfname,".bak");

    oldf = fopen(fname, "r+");
    if (!oldf) return -1;

    newf = fopen(newfname, "w+");

    while (1) {
	int len = fread(data,1,sizeof(data),oldf);
	if (len == 0) break;

	fwrite(data,1,len,newf);
    }

    fclose(oldf);
    fclose(newf);

    return 0;
}
