/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


/*
 * machine depenedent code
 */

#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include <mmsclie.h>
#include <mmclib.h>
#include "common.h"
#include "screen.h"

#include "and_scr.h"
#include "windows.h"
#include "console.h"
#include "menus.h"
#include "clinet.h"

extern void (*wait_events)();

extern int g_BypassLogin;

extern int quit_in_progress;

extern select_info gl_si;

char *KeyMapDirectory = "/usr/local/lib/ezmail/";

static int  net_timeout_sec = 60;
static unsigned send_keepalives = 0;

static int keepalive_timer = 0;
static int checkpoint_timer = 0;

static Boolean g_NoCheckpoints = FALSE;

void help_on_arguments(void);

#define CHECKPOINT_FREQ (5 * 60) /* checkpoint frequency in seconds */

void clear_keepalive_timer(void)
{
  keepalive_timer = get_itimer_value(); /* sync with the timer */
}

#define KEEP_ALIVE_THIS_OFTEN (5 * 60) /* keep alive this often */
void keep_server_alive()
{
  if (((get_itimer_value() - keepalive_timer) < KEEP_ALIVE_THIS_OFTEN) ||
      (is_net_dead())) return;

  send_keepalives = 0;
  RL_do_num_command(SEND_RL, 9, -1L);
}

void mainloop(void)
{
  while(TRUE) {
    (*wait_events)();
    if (gl_si.nfound > 0) {
      if (FD_ISSET(0, &gl_si.read_fd)) {
	while (SCR_Keyboard_io()) {
	  int ch = SCR_GetKey();
	  send_keepalives++;
	  if (WIN_DoKey(ch) == FALSE) MNU_MenuMode(ch);
	}
      }
    }
    if (g_NoCheckpoints == FALSE) {
      if ((get_itimer_value() - checkpoint_timer) >= CHECKPOINT_FREQ) {
	checkpoint_openfiles();
	checkpoint_timer = get_itimer_value();
      }
    }
  }
}

void client_select_info(select_info_pt si)
{
  FD_SET(0,&si->read_fd);
}

void noconn_events()
{
    /*  mn_select_init(&gl_si); */
  client_select_info(&gl_si);

  gl_si.timeout.tv_sec = imin(gl_si.timeout.tv_sec, net_timeout_sec);
  gl_si.nfound=select(FD_SET_SIZE, &gl_si.read_fd,
		      &gl_si.write_fd,&gl_si.except_fd,&gl_si.timeout);

  if(gl_si.nfound<0) {
    if(errno != EINTR) CON_WriteLine("select failed");
  }
  errno=0;
}

void connecting_events()
{
    /*  mn_select_init(&gl_si); */
  client_select_info(&gl_si);
  gl_si.timeout.tv_sec = imin(gl_si.timeout.tv_sec, net_timeout_sec);
  do_events(&gl_si);
}

void connected_events()
{
    memset(&gl_si,0,sizeof(gl_si));
    gl_si.timeout.tv_sec=1;
    /*  mn_select_init(&gl_si); */
    client_select_info(&gl_si);
    gl_si.timeout.tv_sec = imin(gl_si.timeout.tv_sec, net_timeout_sec);
    if (send_keepalives != 0) keep_server_alive();
    do_events(&gl_si);
}

void quit_events()
{
  close_cleanup();

  if (quit_in_progress) { /* if we are quitting */
    if (CON_CanClose()) {
      SCR_RestoreVideoEnv();
      exit(0);
    } else CON_WriteLine(MSG_QuitAborted);
  }
  gl_si.nfound = 0;
  errno = 0;
}

void QuitCleared(void)
{
  wait_events = quit_events;
  gl_si.timeout.tv_sec = 1;
}

void read_args(int argc,char **argv)
{
  char *prefs_file = NIL;
  Boolean init_term = TRUE;

  if (argc >= 2) {
    int i = 1;
    while (i < argc) {
      if (strcmp(argv[i], "-p") == 0) {
	if (i++ >= argc) break;
	prefs_file = argv[i];
      }
      else if (strcmp(argv[i], "-expert") == 0) gNovice = FALSE;
      else if (strcmp(argv[i], "-n") == 0) SCR_UseCharSet(NO_C_SET);
      else if (strcmp(argv[i], "-vt100") == 0) SCR_UseCharSet(VT_C_SET);
      else if (strcmp(argv[i], "-h19") == 0) SCR_UseCharSet(ZT_C_SET);
      else if (strcmp(argv[i], "-b") == 0) WIN_SetMode(W_NOBAN_MODE);
      else if (strcmp(argv[i], "-d") == 0) WIN_SetMode(W_NOBORD_MODE);
      else if (strcmp(argv[i], "-z") == 0) WIN_SetMode(W_ZOOM_ALL);
      else if (strcmp(argv[i], "-h") == 0) SCR_NoHighlight();
      else if (strcmp(argv[i], "-is") == 0) init_term = FALSE;
      else if (strcmp(argv[i], "-c") == 0) g_BypassLogin = FALSE;
      else if (strcmp(argv[i], "-nc")== 0) g_NoCheckpoints = TRUE;
      else if (strcmp(argv[i], "-capture") == 0) gCapturedShell = TRUE;
      else help_on_arguments();
      i++;
    }
  }
  if (init_term) SCR_SendIS();
  /*read_local_prefs(prefs_file);*/ /* if NIL, default will be read */
}

void help_on_arguments(void)
{
  SCR_RestoreVideoEnv();

  printf("Available switches are:\n\n");
  printf("-b       remove extra line in window banners to have more room\n");
  printf("-c       do not bypass the login dialog box for connect\n");
  printf("-d       zoom all and remove borders\n");
  printf("-expert  suppress help message at the bottom of the screen\n");
  printf("-h19     use Heath-19 alternate character set\n");
  printf("-h       do not do any highlights\n");
  printf("-is      do not send initialization string to terminal\n");
#ifdef AFS_ENV
  printf("-l       Allow for Local/Remote operations for menus such as Save/Open/etc.\n");
#endif
  printf("-n       do not use alternate character set\n");
  printf("-nc      do not checkpoint open compose windows\n");
  printf("-p       preferences_file_name\n");
  printf("-vt100   use vt100 alternate character set\n");
  printf("-w       password : This has been disabled due to security problem\n");
  printf("-z       zoom all windows\n");

  exit(0);
}

char *get_machine_config(char *buf,int size)
{
  extern char *getvstr(), *getvstr_cs(), *getenv();
  char *home, *pwd, *term, *termcap;

  if ((home = getenv("HOME")) == 0) home = "No Home";
  if ((pwd  = getenv("PWD"))  == 0) pwd  = "No Pwd";
  if ((term = getenv("TERM")) == 0) term = "Unknown";
  if ((termcap = getenv("TERMCAP")) == 0) termcap = "probably /etc/termcap";
  snprintf(buf, size-1, "%s\nHome: %s\nPwd: %s\nTerm: %s\nTermcap: %s\n\n",
#ifdef IBMPC
	  getvstr()
#else
	  getvstr_cs()
#endif
	  , home, pwd, term, termcap);
  return((char *)(buf + strlen(buf)));
}

/*
 * get some memory, fail if we can't
 */

char *get_deallocable_mem(xx_bytes)
long xx_bytes;
{
  return((char *)malloc(xx_bytes));
}

void free_deallocable_mem(char *xx_addr)
{
  if (xx_addr) free(xx_addr);
}

char *realloc_deallocable_mem(xx_buf, xx_bytes, xx_oldbytes)
char *xx_buf;
long xx_bytes, xx_oldbytes;
{
  return((char *)realloc(xx_buf, xx_bytes));
}

char *c_get_mem(size)
int size;
{
  char *result;
  result = malloc(size);
  if(result==NIL)rl_error("c_get_mem_size %d");
  return result;
}

#if defined(mips) && defined(pmax_3)
char *memset(dst, c, n)
char *dst;
int n, c;
{
  while (n-- > 0) *dst++ = (char)c;
}
#endif /* mips */
