/*
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "mime.h"
#include "attach.h"
#include "stristr.h"

/* An implementation of some handy RFC 2045, 2046, 2047 MIME-related
 * functions for ezimail.
 */



const char HEX_CHARS[]             = "0123456789ABCDEF";
const char RFC2045_SPECIAL_CHARS[] = "@.,;:<>[]\\\"()?/= \t";
const char RFC2060_NIL[] = "NIL";

/* Check whether the given character is safe to send over mail gateways
 * without being clobbered.  EBCDIC machines are still alive and well.
 */
int rfc2045_is_safe_char ( const char c )
{
  return ( ( ( c >= 33 && c <= 126 ) && 
	     ( c != 61 ) && 
	     ( !strchr ( RFC2045_SPECIAL_CHARS, c ) ) ) ? 1 : 0 );
}

/* Is this whitespace chracter safe to send over a mail gateway? */
int rfc2045_is_safe_ws ( const char c ) 
{
  return ( ( c == ' ' || c == '\t' ) ? 1 : 0 );
}

/* Turn a 2-digit hex value into a chracter */
char rfc2045_hex_to_char ( const char * value )
{
  char
    buf[3];
  
  buf[0] = value[0];
  buf[1] = value[1];
  buf[2] = '\0';

  return (char) strtol ( buf, NULL, 16 );
}

/* Turn a charcter into a 2-digit RFC2045 hex number, need 
 * a buffer of at least three digits 
 */
char * rfc2045_char_to_hex ( const char c, char * buf )
{
  buf[0] = HEX_CHARS[c / 16];
  buf[1] = HEX_CHARS[c % 16];
  buf[2] = '\0';
  return buf + 2;
}


/* This function is an RFC2045-compliant quoted-printable
 * encoder.  It takes a source buffer (null terminated) and 
 * destination buffer, along with the destination buffer's
 * length.
 * 
 * If there is enough buffer space, the length of the destination
 * buffer is returned, otherwise -1 is returned and the function
 * should be called with a larger buffer
 *
 */
int rfc2045_to_mime_quoted_printable ( const char * s,
				       char       * d,
				       int          outLen )
{
  char 
    * inP,
    * outP;

  int
    linePos = 0,
    outPos  = 0;
  
  inP  = (char*)s;
  outP = (char*)d;

  while ( *inP != '\0' ) { 
    
    if ( outPos + 1 >= outLen )
      return -1;

    if ( rfc2045_is_safe_char ( *inP ) ||
	 ( linePos < RFC2045_MAX_LINE_LEN && rfc2045_is_safe_ws ( *inP ) ) ) { 
      *outP++ = *inP++, outPos++;
      linePos++;

      if ( linePos >= RFC2045_MAX_LINE_LEN ) {
	if ( outPos + 3 >= outLen ) 
	  return -1;
	*outP++ = '=', outPos++;
	*outP++ = '\r', outPos++;
	*outP++ = '\n', outPos++;
	linePos = 0;
      }
      
    } else { 

      if ( outPos + 3 >= outLen )
	return -1;

      /* Need to encode it as a =XX character */
      if ( linePos >= ( RFC2045_MAX_LINE_LEN - 2 ) ) {
	*outP++ = '=', outPos++;
	*outP++ = '\r', outPos++;
	*outP++ = '\n', outPos++;
	linePos = 0;
      }
      
      *outP++ = '=', outPos++;
      outP = rfc2045_char_to_hex ( *inP++, outP ), outPos += 2;
      linePos += 3;
    }
  }

  *outP = '\0';
  return outPos;
}

/* Convert a quoted-printable buffer into a normal unix 
 * text buffer that is suitable for editing.
 */
int rfc2045_from_mime_quoted_printable ( const char * in, 
					 char * out,
					 int    outLen,
                                         const int removeReturns )
{
  char 
    * outP,
    * inP;
  
  int
    outPos = 0;
  
  inP  = (char*)in;
  outP = (char*)out;
  
  if ( in == NULL || out == NULL || outLen <= 0 ) 
    return -1;
  
  while ( outPos < outLen && *inP != '\0' ) { 
    while ( *inP != '\0' && *inP != '=' && outPos < outLen && *inP != '\r' ) {
      *outP = *inP;
      outP++; inP++; outPos++;
    }

    /* Get rid of trailing whitespace */
    if ( *inP == '\r' ) { 
      char 
	* inP2 = inP - 1;

      while ( isspace ( *inP2 ) )
	inP2--, outP--, outPos--;

      inP += 1;
      if ( !removeReturns ) { 
        *outP = '\r';
	outP++;
        outPos++;
      }
      *outP = '\n';
      outP++;
      outPos++;
    }

    if ( *inP == '=' ) {
      inP++;
      if ( *(inP) == '\0'|| *(inP+1) == '\0' ) {
	*outP = '\0';
	return outPos;
      }
      
      /* Check for an = at the end of the line */
      if ( !(*inP == '\r' && *(inP+1) == '\n') ) {
	*outP++ = rfc2045_hex_to_char ( inP );
	outPos++;
      }
      inP += 2;
    }
  }
  *outP = '\0';

  return outPos;
}

const char * rfc2045_mimetype_to_string ( const int mimeType )
{
  switch ( mimeType ) { 
  case MIME_TYPE_TEXT:        return "text";
  case MIME_TYPE_IMAGE:       return "image";      
  case MIME_TYPE_AUDIO:       return "audio";      
  case MIME_TYPE_VIDEO:       return "video";      
  case MIME_TYPE_APPLICATION: return "application";
  case MIME_TYPE_EXTENSION:   return "X-something";
  case MIME_TYPE_MESSAGE:     return "message";  
  case MIME_TYPE_MULTIPART:   return "multipart";
  default:                    return "X-unknown";
  } 


}


char * rfc2045_mime_headers ( int    type,
			      char * subtype,
			      char * charset,
			      char * encoding,
                              char * filename,
                              RFC2183_DISPOSITION_TYPE rfc2183_disposition_type,
                              char * description,
                              char * boundary,
			      char * buffer,
			      int    bufLen )
{
  int 
    newLen = 0;
  
  char 
    * typeStr;
  
  /* These are reasonable defaults for very portable ASCII text */
  switch ( type ) { 
  case MIME_TYPE_TEXT:        typeStr = "text";        break;
  case MIME_TYPE_IMAGE:       typeStr = "image";       break;
  case MIME_TYPE_AUDIO:       typeStr = "audio";       break;
  case MIME_TYPE_VIDEO:       typeStr = "video";       break;
  case MIME_TYPE_APPLICATION: typeStr = "application"; break;
  case MIME_TYPE_EXTENSION:   typeStr = "X-something"; break;
  case MIME_TYPE_MESSAGE:     typeStr = "message";     break;
  case MIME_TYPE_MULTIPART:   typeStr = "multipart";   break;
  default:                    typeStr = "X-unknown";   break;
  } 

  if ( subtype == NULL )  subtype = "plain";
  if ( encoding == NULL ) encoding = "7bit";
  
  newLen += snprintf ( buffer + newLen,
                       bufLen - newLen,
                      "Content-Type: %s/%s",
                      typeStr,
                      subtype );

  if ( boundary != NULL ) { 
    newLen += snprintf ( buffer + newLen, 
                         bufLen - newLen,
                        "; boundary=\"%s\"",
                        boundary );
  }

  if ( charset != NULL ) { 
    newLen += snprintf ( buffer + newLen,
                         bufLen - newLen,
                         "; charset=\"%s\"; ",
                         charset );
  }

  if ( filename != NULL ) { 
    /* Pine adds this header.  We might as well. */
    newLen += snprintf ( buffer + newLen,
                         bufLen - newLen,
                         "; name=\"%s\"; ",
                         filename );
  }
  
  newLen += snprintf ( buffer + newLen,
                       bufLen - newLen,
                       "\r\n" );

  newLen += rfc2183_mime_content_disposition ( filename,
                                               rfc2183_disposition_type,
                                               buffer + newLen,
                                               bufLen - newLen );

  if ( description != NULL ) {  
    newLen += snprintf ( buffer + newLen, bufLen - newLen,
                         "Content-Description: \"%s\"\r\n",
                         description );
  }

  newLen += snprintf ( buffer + newLen, 
                       bufLen - newLen,
                      "Content-Transfer-Encoding: %s\r\n",
                      encoding );

  newLen += snprintf ( buffer + newLen,
                       bufLen - newLen,
                       "\r\n" );

  return ( buffer + newLen );
}

/* Add an RFC2183 Content-Disposition header to the message, if enough
 * arguments have been given
 */
int    rfc2183_mime_content_disposition ( char  * filename,
                                          RFC2183_DISPOSITION_TYPE type,
                                          char  * buffer,
                                          int     bufLen )
{
  int
    bufUsed = 0;
  
  char
    * typeStr = NULL;

  
  switch ( type ) { 
  case RFC2183_INLINE:     typeStr = "inline";     break;
  case RFC2183_ATTACHMENT: typeStr = "attachment"; break;
  case RFC2183_UNKNOWN:  
  default:
    return 0;
  }

  bufUsed += snprintf ( buffer + bufUsed,
                        bufLen - bufUsed,
                        "Content-Disposition: %s",
                        typeStr );
  if ( filename ) {
    bufUsed += snprintf ( buffer + bufUsed,
                          bufLen - bufUsed,
                          "; filename=\"%s\";",
                          filename );
  }
    
  bufUsed += snprintf ( buffer + bufUsed,
                        bufLen - bufUsed,
                        "\r\n" );
  
  return bufUsed;
}

/* Decode the RFC 2060 FETCH x BODYSTRUCTURE response.
 * Mark Crispin must be removed from this planet for having written
 * such an incomprehensible description on this in 7.4.2 FETCH Response.
 *
 * A BNF specification would have made his life much easier and made
 * me much happier.  The guy should be placed in a remedial English 
 * class.  Phew.  I feel better now.
 */

/* Get the next string token */
#define MAX_STRING_SZ 8192

#define WARP_WS(BS) do { while isspace ( *BS ) BS++; } while ( 0 )

static char * malloc_copy_string ( char ** buf, const char * data )
{
  *buf = malloc ( strlen ( data ) + 1 );
  strcpy ( *buf, data );
  return *buf;
}

static char * next_string_token ( char * bs, char ** token )
{
  int
    len = 0;
  char
    buf[MAX_STRING_SZ] = { 0 },
    *bufp;
    
    //printf ( "Processing: %s\n", bs );
    
    WARP_WS ( bs );
    
    bufp = buf;
    
    /* Get the NIL arguments */
    if ( !strncasecmp ( RFC2060_NIL, bs, strlen ( RFC2060_NIL ) ) ) { 
      malloc_copy_string ( token, RFC2060_NIL );
      printf ( "Found string token as %s\n", *token );
      return ( bs + strlen ( RFC2060_NIL ) );
    }
    
    if ( *bs++ != '\"' ) {
      printf ( "Failed to find quotes\n" );
      return NULL;
    }
    
    while ( *bs != '\0' && 
	    *bs != '\"' && len < MAX_STRING_SZ - 1 ) 
      *bufp++ = *bs++, len++;
    
    printf ( "Found string token: %s (len=%d)\n", ( bufp  - len ), len );
    if ( *bs == '\"' ) {
      malloc_copy_string ( token, buf );
      return ++bs;
    }
    
    return NULL;
}

/* Get the next integer token */
static char * next_integer_token ( char * bs, int * token )
{
  while ( isspace ( *bs ) )
    bs++;

  if ( !strncasecmp ( RFC2060_NIL, bs, strlen ( RFC2060_NIL ) ) ) { 
    bs += strlen ( RFC2060_NIL );
    *token = -1;
    return bs;
  }
  
  if ( !isdigit ( *bs ) )
    return NULL;
  
  *token = strtol ( bs, &bs, 10 );
  
  return bs;
}

/* Slurp up an entire parameter list in parens, in case we don't
 * know how to parse it.
 */
static char * next_paren_list_token ( char * bs, char ** token )
{
  int 
    parenCounter = 0,
    tokenPos = 0,
    inQuotes = 0;

  char
    buf[MAX_STRING_SZ] = { 0 },
    *bufp;

  /* ( "foo" "bar" ... ), allow nested parens and parens in quotes */
  while ( isspace ( *bs ) )
    bs++;
  
  if ( *bs != '(' )
    return NULL;

  bufp = buf;
  *bufp++ = *bs++, tokenPos++;
  parenCounter++;

  while ( *bs != '\0' &&  tokenPos < MAX_STRING_SZ && parenCounter > 0 ) { 
    
    switch ( *bs ) { 
    case '(':
      if ( !inQuotes ) parenCounter++;
      break;
    case ')':
      if ( !inQuotes ) parenCounter--;
      break;
    case '\"':
      inQuotes = !inQuotes;
      break;
    }
    *bufp++ = *bs++, tokenPos++;
  }
  printf ( "Found paren Token: %s, len = %d\n", bufp - tokenPos, tokenPos );
  if ( parenCounter > 0 )
    return NULL;

  malloc_copy_string ( token, buf );

  return bs;
}

static int get_parameter_type ( const char * name )
{
  if ( name == NULL ) 
    return MIME_PARAMETER_TYPE_INVALID;

  /* JCS - I only know one type for now */
  if ( !strncasecmp ( "NAME", name, strlen ( "NAME" ) ) ) return MIME_PARAMETER_TYPE_NAME;

  return MIME_PARAMETER_TYPE_UNKNOWN;

}

static char * get_parameter_pair ( char * bs, MimeAttachmentParameter ** param )
{
  char 
    * name,
    * value;

  WARP_WS ( bs );

  if ( *bs != '(' ) 
    return NULL;  
  bs++;

  bs = next_string_token ( bs, &name );
  if ( bs == NULL ) return NULL;
  bs = next_string_token ( bs, &value );
  if ( bs == NULL ) return NULL;

  *param = alloc_parm();
  (*param)->name  = name;
  (*param)->value = value;
  (*param)->type  = get_parameter_type ( name );

  WARP_WS ( bs );
  if ( *bs != ')' )
    return NULL;

  
  
  return ++bs;
}

/* Same as above, except also take a single string token */
static char * next_paren_list_or_pair_string_token ( char * bs, char ** token )
{
  char
    *tok1,
    *tok2;

  while ( isspace ( *bs ) )
    bs++;

  printf ( "Processing: %s\n", bs );
  if ( *bs == '(' ) 
    return next_paren_list_token ( bs, token );
  
  if ( next_string_token ( bs, &tok1 ) != NULL ) {
    bs = next_string_token ( bs, &tok1 );
    if ( bs == NULL ) return NULL;
    bs = next_string_token ( bs, &tok2 ); 
    if ( bs == NULL ) return NULL;
    if ( strncasecmp ( tok1, RFC2060_NIL, strlen ( RFC2060_NIL ) ) ) {
      *token = malloc ( 6 + strlen ( tok1 ) + strlen ( tok2 ) );
      sprintf ( *token, "\"%s\" \"%s\"", tok1, tok2 );
    } else { 
      *token = malloc ( 6 + strlen ( tok1 ) + strlen ( tok2 ) );
      sprintf ( *token, "%s %s", tok1, tok2 );
    }
    printf ( "Pair token: %s\n", *token );
    return bs;
  } else
    return NULL;
}

int rfc2060_mime_type ( char * str )
{ 
  if ( !strncasecmp ( "TEXT",        str, 4  ) ) return MIME_TYPE_TEXT;
  if ( !strncasecmp ( "IMAGE",       str, 5  ) ) return MIME_TYPE_IMAGE;
  if ( !strncasecmp ( "AUDIO",       str, 5  ) ) return MIME_TYPE_AUDIO;
  if ( !strncasecmp ( "VIDEO",       str, 5  ) ) return MIME_TYPE_VIDEO;
  if ( !strncasecmp ( "APPLICATION", str, 11 ) ) return MIME_TYPE_APPLICATION;
  if ( !strncasecmp ( "X-",          str, 2  ) ) return MIME_TYPE_EXTENSION;
  if ( !strncasecmp ( "MESSAGE",     str, 7  ) ) return MIME_TYPE_MESSAGE;
  if ( !strncasecmp ( "MULTIPART",   str, 9  ) ) return MIME_TYPE_MULTIPART;

  return MIME_TYPE_INVALID;
}

/* Retreive the integer constant version of the first field 'type'
 * in the body structure.  Note that this blows away the extension type,
 * but we don't know much about how to support that now anyway...
 */
char * rfc2060_get_type ( char * bs, int * type )
{
  char 
    *typeStr;
  
  bs = next_string_token ( bs, &typeStr );
  
  *type = MIME_TYPE_INVALID;
  
  if ( bs == NULL ) 
    return NULL;

  *type = rfc2060_mime_type ( typeStr );

  return bs;
}

char * rfc2060_get_subtype ( char * bs, char ** subtype )
{
  return next_string_token ( bs, subtype );
}

char * rfc2060_get_body_parameter ( char * bs, char ** params )
{
  return next_paren_list_token ( bs, params );
}

char * rfc2060_get_body_id ( char * bs, char ** id )
{
  return next_string_token ( bs, id );
}

char * rfc2060_get_body_description ( char * bs, char ** description )
{
  return next_string_token ( bs, description );
}

char * rfc2060_get_body_encoding ( char * bs, char ** encoding )
{
  return next_string_token ( bs, encoding );
}

char * rfc2060_get_body_size ( char * bs, int * octets, int * lines )
{
  bs = next_integer_token ( bs, octets );
  if ( bs == NULL ) return NULL;
  
  bs = next_integer_token ( bs, lines );
  if ( bs == NULL ) return NULL;

  return bs;
}

char * rfc2060_get_body_language ( char * bs, char ** language )
{
  return next_string_token ( bs, language );
}

char * rfc2060_get_text_body_size ( char * bs, int * lines )
{
  return next_integer_token ( bs, lines );
}

char * rfc2060_get_extension_fields ( char *  bs, 
				      char ** parameter,
				      char ** disposition,
				      char ** language )
{
  while ( isspace ( *bs ) )
    bs++;
  if ( *bs == ')' )
    return ++bs;

  bs = next_paren_list_or_pair_string_token ( bs, parameter );
  if ( bs == NULL ) return NULL;

  bs = next_string_token ( bs, disposition );
  if ( bs == NULL ) return NULL;
  
  bs = next_string_token ( bs, language );
  if ( bs == NULL ) return NULL;
  
  return bs;
}

MimeAttachmentParameter * alloc_parm ( void )
{
  MimeAttachmentParameter
    * map;
  
  map = malloc ( sizeof(MimeAttachmentParameter) );
  if ( map == NULL )
    return NULL;
  bzero ( map, sizeof(MimeAttachmentList) );
  return map;
}

MimeAttachmentList * alloc_mal ( void )
{
  MimeAttachmentList 
    * mal;
  mal = malloc ( sizeof(MimeAttachmentList) );
  if ( mal == NULL ) 
    return NULL;
  bzero ( mal, sizeof(MimeAttachmentList) );
  return mal;
}

MimeAttachment * alloc_ma ( void )
{
  MimeAttachment
    * mal;
  mal = malloc ( sizeof(MimeAttachment) );
  if ( mal == NULL ) 
    return NULL;
  bzero ( mal, sizeof(MimeAttachment) );
  return mal;
}

/* The discrete media types are layed out as:
 *  "type" "subtype" (parameter list) "id" "description" 
 *  "encoding" octet_size line_size 
 */

char * rfc2060_decode_discrete ( char * bs, 
				 MimeAttachmentList * mal,
				 int type )
{
  printf ( "Decoding discrete...\n" );

  /* Allocate and decode the current discrete type */
  mal->value = alloc_ma();
  if ( mal->value == NULL ) return NULL;

  mal->value->type = type;
  bs = rfc2060_get_subtype ( bs, &mal->value->subtype );
  if ( bs == NULL ) return NULL;

  bs = rfc2060_get_body_parameter ( bs, &mal->value->parameter );
  if ( bs == NULL ) return NULL;

  get_parameter_pair ( mal->value->parameter, &mal->value->param );
  
  bs = rfc2060_get_body_id ( bs, &mal->value->id );
  if ( bs == NULL ) return NULL;
  
  bs = rfc2060_get_body_description ( bs, &mal->value->description );
  if ( bs == NULL ) return NULL;
  
  bs = rfc2060_get_body_encoding ( bs, &mal->value->encoding );
  if ( bs == NULL ) return NULL;

  bs = rfc2060_get_body_size ( bs,
			       &mal->value->sizeOctets,
			       &mal->value->sizeLines );
  if ( bs == NULL ) return NULL;

  bs = rfc2060_get_extension_fields ( bs, 
				     &mal->value->ext_parameter,
				     &mal->value->ext_disposition,
				     &mal->value->ext_language );
  if ( bs == NULL ) return NULL;

  while ( isspace ( *bs ) ) 
    bs++;

  if ( *bs != ')' )
    return NULL;
  
  bs++;
  
  return bs;  
}

/* Multipart looks something like:
 * "type" "subtype" (parameter list) (disposition list) (language)|"language"
 */
char * rfc2060_decode_multipart ( char * bs,
				  MimeAttachmentList * mal )
{
  printf ( "Decoding multipart\n" );

  mal->value = alloc_ma();
  if ( mal->value == NULL ) return NULL;
  mal->value->type = MIME_TYPE_MULTIPART;

  bs = rfc2060_get_subtype ( bs, &mal->value->subtype );
  if ( bs == NULL ) return NULL;

  bs = rfc2060_get_body_parameter ( bs, &mal->value->ext_parameter );
  if ( bs == NULL ) return NULL;

  bs = rfc2060_get_body_parameter ( bs, &mal->value->ext_disposition );
  if ( bs == NULL ) return NULL;

  bs = next_paren_list_or_pair_string_token ( bs, &mal->value->ext_language );
  if ( bs == NULL ) return NULL;

  while ( isspace ( *bs ) )
    bs++;
  
  if ( *bs++ == '(' ) {
    mal->leaf = alloc_mal();
    return rfc2060_decode_structure_list ( bs, mal->leaf );
  }

  return NULL;
}

char * rfc2060_decode_message ( char * bs,
				  MimeAttachmentList * mal )
{
  //  char 
  //  MD5[STR255LEN];

  printf ( "Decoding message\n" );

  mal->value = alloc_ma();
  if ( mal->value == NULL ) return NULL;

  mal->value->type = MIME_TYPE_MESSAGE;
  bs = rfc2060_get_subtype ( bs, &mal->value->subtype );
  if ( bs == NULL ) return NULL;

  bs = rfc2060_get_body_parameter ( bs, &mal->value->parameter );
  if ( bs == NULL ) return NULL;
  get_parameter_pair ( mal->value->parameter, &mal->value->param );
  
  bs = rfc2060_get_body_id ( bs, &mal->value->id );
  if ( bs == NULL ) return NULL;
  
  bs = rfc2060_get_body_description ( bs, &mal->value->description );
  if ( bs == NULL ) return NULL;
  
  bs = rfc2060_get_body_encoding ( bs, &mal->value->encoding );
  if ( bs == NULL ) return NULL;

  bs = next_integer_token ( bs, 
			    &mal->value->sizeLines );
  if ( bs == NULL ) return NULL;

  bs = next_paren_list_or_pair_string_token ( bs, 
					      &mal->value->ext_disposition );
  if ( bs == NULL ) return NULL;
  
  while ( isspace ( *bs ) )
    bs++;
  
  if ( *bs++ == '(' ) {
    mal->leaf = alloc_mal();
    return rfc2060_decode_structure_list ( bs, mal->leaf );
  }

  return bs;
}
				  
/* Go horizontally across the bodystructure... and recurse in
 *  ( ... ) ( ... ) ( ... ) 
 */
char * rfc2060_decode_structure_list ( char * bs, 
				       MimeAttachmentList * mal )
{
  int
    type;

  MimeAttachmentList 
    * currNode = mal;
  
  while ( isspace ( *bs ) )
    bs++;
  
  while ( *bs++ == '(' ) { 

    while ( isspace ( *bs ) )
      bs++;
    
      printf ( "Linear, processing.. currnode = %x\n", (int)currNode );
      
      if ( currNode->next ) 
	currNode = currNode->next;

      /* This is an actual mime attachment to decode, get the type */
      bs = rfc2060_get_type ( bs, &type );
      if ( type == MIME_TYPE_INVALID ) { 
	printf ( "Could not determine mime type\n" );
	return NULL;
      }
      switch ( type ) { 
      case MIME_TYPE_MESSAGE:
	printf ( "This is a message\n" );
	bs = rfc2060_decode_message ( bs, currNode );
	break;
      case MIME_TYPE_MULTIPART:
	printf ( "This is a multipart\n" );
	bs = rfc2060_decode_multipart ( bs, currNode );
	break;
      default:
	printf ( "This is a discrete\n" );
	bs = rfc2060_decode_discrete ( bs, currNode, type );
	break;
      }
      if ( bs == NULL ) return NULL;
      currNode->next = alloc_mal();
      if ( currNode->next == NULL )
	return NULL;

      while ( isspace ( *bs ) )
	bs++;
      printf ( "Done with one part, this is left: %s\n*********************\n", bs  );
  }
  
  if ( currNode->next ) {
    /* Reclaim memory for an unused node */
    free ( currNode->next );
    currNode->next = NULL;
  }
  
  return bs;
}
		      
MimeAttachmentList * rfc2060_decode_bodystructure ( char * bs )
{
  MimeAttachmentList
    * mal;

  mal = alloc_mal();
  if ( mal == NULL ) 
    return NULL;

  
  while ( isspace ( *bs ) )
    bs++;
  
  if ( *bs != '(' ) {
    printf ( "Bodystructure not found\n" );
    return NULL;
  }
  
  bs++;

  bs = rfc2060_decode_structure_list ( bs,
				       mal );

  return mal;  
}

void mal_print ( MimeAttachmentList * mal )
{
  if ( mal->leaf ) {
    printf ( "** Recursing...\n" );
    mal_print ( mal->leaf );
    printf ( "** Popping recursion\n" );
  }

  if ( mal->value ) { 
    printf ( "Type: %d\nsubtype: %s\nparameter: %s\nid: %s\ndesc: %s\nencoding: %s\nsizebytes: %d\nsizelines: %d\next_param: %s\next_dispos: %s\next_lang: %s\n",
	     mal->value->type,
	     mal->value->subtype,
	     mal->value->parameter,
	     mal->value->id,
	     mal->value->description,
	     mal->value->encoding,
	     mal->value->sizeOctets,
	     mal->value->sizeLines,
	     mal->value->ext_parameter,
	     mal->value->ext_disposition,
	     mal->value->ext_language );
    if ( mal->value->param ) { 
      printf ( " Param name = %s value = %s type = %d\n", 
	       mal->value->param->name,
	       mal->value->param->value,
	       mal->value->param->type );
    }
  } else { 
    printf ( "NO VALUE!\n" );
  }
  

  if ( mal->next ) { 
    printf ( "*********** NEXT IN LIST\n" );
    mal_print ( mal->next );
  }

}

/* Generate a MIME RFC 2046 boundary token, as
 * recommended in the RFC (although this implementation will
 * only generate alphanumeric tokens).
 */
#define LETTERS 26
#define NUMBERS 10
char * rfc2046_mime_boundary ( char * buf, 
                               int    len )
{
  int
    num,
    i;

  /* Need to have enough space for the token
   */
  if ( len < 1 || buf == NULL ) { 
    return NULL;
  }

  buf[len-1] = '\0';
  
  for ( i = 0; i < len - 1; i++ ) { 
    /* Apparently, some implemntations of rand() will be less
     * random in the lower bits than middle bits, so grab from the 
     * middle. <snort>
     */
    num = ( rand() >> 4 ) % (2 * LETTERS + NUMBERS);
    if ( num < 0 )
      num = -num;

    if ( num < NUMBERS )
      buf[i] = (char)('0' + num);
    else if ( num < ( LETTERS + NUMBERS ) ) 
      buf[i] = (char)('A' + num - NUMBERS );
    else
      buf[i] = (char)('a' + num - NUMBERS - LETTERS);
  }

  return buf;
}

/* Generate a MIME version header */
char * rfc2046_mime_version_header ( char * buf,
                                     int    len )
{
  snprintf ( buf, len, "Mime-Version: 1.0\r\n" );
  return buf;
}

char * get_unquoted_string ( char * strbuf,
                             char * outbuf,
                             const int outbuflen )
{
  int 
    i;
  
  while ( isspace ( *strbuf ) ) strbuf++;
  
  if ( *strbuf == '"' ) { 
    
    strbuf++;
    i = 0;
    while ( strbuf[i] != '"' && i < ( outbuflen - 1 ) ) {
      outbuf[i] = strbuf[i];
      i++;
    }
    
  } else { 
    i = 0;
    while ( !isspace ( strbuf[i] ) && i < ( outbuflen - 1 ) ) {
      outbuf[i] = strbuf[i];
      i++;
    }
  }

  return strbuf;  
}

/* Search for the following fields in the return value from the MIME
 * headers returned from an RFC2060 machine, looking for RFC2045 and
 * RFC2183 headers
 */
int rfc2060_get_mime_headers ( const char *  buf,  /* The buffer */
                               const int     len,  /* The length of buf */
                               RFC2060_MimeHeaders * headers )
                               
{
  char 
    * ptr;

  int
    i;

  bzero ( headers, sizeof(RFC2060_MimeHeaders) );
  headers->mimeType    = MIME_TYPE_TEXT;
  headers->disposition = RFC2183_UNKNOWN;
  strcpy ( headers->mimeSubType, "plain" );
  strcpy ( headers->encoding,    "7bit" );

  /* Make sure we have a null-terminated buffer */
  if ( buf[len] != '\0' ) { 
    return -1;
  } 
  
  /* Extract useful information from the buffer */
  ptr = stristr ( buf, CONTENT_TYPE );
  if ( ptr != NULL ) { 
    
    ptr += ( strlen ( CONTENT_TYPE ) + 1 );

    while ( isspace ( *ptr ) ) ptr++;
    
    headers->mimeType = rfc2060_mime_type ( ptr );
    
    while ( *ptr && *ptr != '/' ) ptr++;

    if ( *ptr != '/' ) { 
      headers->mimeType = MIME_TYPE_INVALID;
    } else { 

      /* Locate the mime subtype */
      ptr++;
      
      i = 0; 
      while ( i < ( STR255LEN - 1 ) && rfc2045_is_safe_char ( *ptr )  ) { 
        headers->mimeSubType[i] = *ptr;
        i++; ptr++;
      }
      headers->mimeSubType[i] = '\0';
    }
    
    while ( isspace ( *ptr ) || *ptr == ';' ) ptr++;
    if ( !strncasecmp ( ptr, CONTENT_TYPE_NAME,
                                   strlen ( CONTENT_TYPE_NAME ) ) ) { 
      ptr += strlen ( CONTENT_TYPE_NAME );
      ptr = get_unquoted_string ( ptr, headers->name, STR255LEN ); 
    }
  }

  ptr = stristr ( buf, CONTENT_DISPOSITION );
  if ( ptr != NULL ) {

    ptr += ( strlen ( CONTENT_DISPOSITION ) + 1 );
    
    while ( isspace ( *ptr ) ) ptr++;
    
    if ( strncasecmp ( DISPOSITION_ATTACHMENT, ptr, strlen ( DISPOSITION_ATTACHMENT ) ) == 0 ) { 
      
      headers->disposition = RFC2183_ATTACHMENT;

      ptr += strlen ( DISPOSITION_ATTACHMENT );
      while ( isspace ( *ptr ) || *ptr == ';' ) ptr++;
      
      if ( !strncasecmp ( ptr, DISPOSITION_FILENAME, 
                          strlen ( DISPOSITION_FILENAME ) ) ) { 
        
        ptr += strlen ( DISPOSITION_FILENAME );
        ptr = get_unquoted_string ( ptr, headers->filename, STR255LEN );
        
      }

    } else if ( strncasecmp ( ptr, DISPOSITION_INLINE, strlen ( DISPOSITION_INLINE ) ) == 0 ) { 
      
      headers->disposition = RFC2183_INLINE;

      ptr += strlen ( DISPOSITION_INLINE );
      while ( isspace ( *ptr ) || *ptr == ';' ) ptr++;
      
      if ( !strncasecmp ( ptr, DISPOSITION_FILENAME, 
                          strlen ( DISPOSITION_FILENAME ) ) ) { 
        
        ptr += strlen ( DISPOSITION_FILENAME );
        ptr = get_unquoted_string ( ptr, headers->filename, STR255LEN );
      }
    }

  }

  /* Find Content-Transfer-Encoding */
  ptr = stristr ( buf, CONTENT_TRANSFER_ENCODING );
  if ( ptr != NULL ) { 

    ptr += ( strlen ( CONTENT_TRANSFER_ENCODING ) + 1 );

    while ( isspace ( *ptr ) ) ptr++;
    
    i = 0;      
    while ( i < ( STR255LEN - 1 ) && rfc2045_is_safe_char ( ptr[i] )  ) { 
      headers->encoding[i] = ptr[i];
      i++;
    }

  }
    
  return 0;
}

#if 0
char inbuffer[3000] = { 0 };
char outbuffer[30000] = { 0 };

int main ( int argc, char ** argv ) 
{
  char * in;
  int i;
  MimeAttachmentList 
    * mal;

  in = inbuffer;
  while ( ( *in = getchar() ) != EOF )
    in++;
  *in = 0;

  printf ( "I read: %d bytes\n", strlen ( inbuffer ) );
  i = rfc2045_from_mime_quoted_printable ( inbuffer,
                                           outbuffer,
                                           sizeof(outbuffer),
                                           FALSE );
  printf ( "Decoded: %d\n%s", i, outbuffer );
  // mal = rfc2060_decode_bodystructure ( inbuffer );
  

  //  mal_print ( mal );
  return 0;
  
}
#endif 

