/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * header file for handling text buffers
 * ie: console buffer, motd buffer, etc.
 */

#ifndef TEXT_H
#define TEXT_H

#include <stdlib.h>
#include <stdio.h>
#include <sys/param.h>
#include <sys/types.h>


#include "list.h"
#include "attach.h"

#define LINE_SIZE 74      /* size of a line while editing */

#define EDT_F_CRLF 0x01   /* line has cr-lf */

typedef struct scr_lines_R {
  long char_pos;
  int  len, row;
  int  flags;
} SCR_LINES, *SCR_LINES_pt;

typedef struct buf_struct {
/*
  These are the fields associated with the text
*/
  int  bsize;          /* buffer size */
  long fsize;          /* file size */
  long offset;         /* Current offset into data */
  long ndata;          /* new data offset, not yet displayed */
  long flags;          /* flags associated with this file */
  char *buf;           /* buffer allocated for this file */
  FILE *stream;        /* stream ptr for this file */
  long mark_start, mark_end, lastmark;
  char lname[MAXPATHLEN]; /* local filename for this file */
#ifndef IBMPC
  long ckpt_count;
  char ckpt_file[MAXPATHLEN];
#endif
/*
  These fields are asscociated with displaying the text
*/
  WINLIST *awp;   /* pointer to window for this text view */
  SCR_LINES_pt scr_lines;
  char ban_text[MAX_BAN+1];
/*
  These fields are associated with editing the text
*/
  long char_pos;          /* insertion point */
  long eop_pos, bop_pos;  /* end/begin of dispaly offsets */
  long hdr_end;           /* headers end at this offset */
  long modify_count;      /* modify count */
  int  x_cursor, y_cursor;
  long gap_start, gap_size;

  /* These fields are associated with file attachments that are 
   * added to documents that are going to be sent
   */
  long attachmentCount;
  Attachment_t attachments[MAX_ATTACHMENTS];
} TEXT_BUF, *TEXT_BUF_pt;

typedef struct text_buf_list_struct {
  TEXT_BUF *a_tbuf;
  struct text_buf_list_struct *prev;
  struct text_buf_list_struct *next;
} TEXT_VIEW, *TEXT_VIEW_pt;

extern TEXT_VIEW *TXT_initialize();

#define TF_CLOSE    0x000001 /* Can close this text view */
#define TF_SCROLL   0x000002 /* Scroll on new data */
#define TF_FREE     0x000004 /* Free on close */
#define TF_TRIM     0x000008 /* Trim buffer if overflow */
#define TF_CENTER   0x000010 /* Center the banner text */
#define TF_CHANGED  0x000020 /* Document has changed */
#define TF_DOCUMENT 0x000040 /* Text view is document */
#define TF_EDITING  0x000080 /* The document is being edited */
#define TF_EDIT     0x000100 /* text view is editable */
#define TF_REMOVE   0x000200 /* Remove file upon close */
#define TF_ISOPEN   0x000400 /* The file is open */
#define TF_OFFDSK   0x000800 /* Page off disk */
#define TF_FLOAD    0x001000 /* load the file into view */
#define TF_LOCKED   0x002000 /* this text view is locked for now */
#define TF_CRBLOW   0x004000 /* Carriage-return closes window */
#define TF_EXACT    0x008000 /* use exact size, not fixed allocation */
#define TF_AFSFILE  0x010000 /* This is an AFS file window */
#define TF_SENT     0x020000 /* JCS - document has been sent */

#define TF_SET_CLOSE(xx)    ((xx) |= TF_CLOSE)
#define TF_SET_SCROLL(xx)   ((xx) |= TF_SCROLL)
#define TF_SET_FREE(xx)     ((xx) |= TF_FREE)
#define TF_SET_TRIM(xx)     ((xx) |= TF_TRIM)
#define TF_SET_CENTER(xx)   ((xx) |= TF_CENTER)
#define TF_SET_CHANGED(xx)  ((xx) |= TF_CHANGED)
#define TF_SET_DOCUMENT(xx) ((xx) |= TF_DOCUMENT)
#define TF_SET_EDITING(xx)  ((xx) |= TF_EDITING)
#define TF_SET_EDIT(xx)     ((xx) |= TF_EDIT)
#define TF_SET_REMOVE(xx)   ((xx) |= TF_REMOVE)
#define TF_SET_ISOPEN(xx)   ((xx) |= TF_ISOPEN)
#define TF_SET_OFFDSK(xx)   ((xx) |= TF_OFFDSK)
#define TF_SET_FLOAD(xx)    ((xx) |= TF_FLOAD)
#define TF_SET_LOCKED(xx)   ((xx) |= TF_LOCKED)
#define TF_SET_AFSFILE(xx)  ((xx) |= TF_AFSFILE)
#define TF_SET_SENT(xx)     ((xx) |= TF_SENT)

/* 
  for the PC, page off of disk file. For larger machines use bigger buffers.
  paging is done if TF_OFFDSK is set in the flags field.
  if FIXED_ALLOCATION is defined, we allocate full PAGEBALE_BUFSIZE in
  one shot to minimize memory fragmentation.
  Note: Make sure that the size of these buffers are large enough so we
  can at least page one at a time. ie: your max screen size.
*/

#ifdef IBMPC
#define PAGEABLE_BUFSIZE 32766  /* use 32K buffere for PC */
#undef FIXED_ALLOCATION        /* allocate at 1 shot */
#else
#define PAGEABLE_BUFSIZE 32767 /* use 32K buffers for larger systems */
#undef FIXED_ALLOCATION       /* allocate at 1 shot */
#endif

extern long EDT_find_last_eol(), EDT_find_next_eol();

int TXT_GetData(TEXT_BUF *tbuf,
		char *buf,
		int len,
		long offset);

void TXT_Edit(void);

void TXT_OpenBlurb(char *filename, char *title);

void TXT_ExtEdit(void);

void TXT_CopyText(void);

int TXT_PrintWindow(void);

void TXT_SetupMenus(int net_state);

int TXT_DoMenuCommand(int m_id,char *str);

void TXT_BringNextDocToFront(void);

int TXT_FormatText(char *text,int len, int width);

int TXT_WriteFile(TEXT_BUF *atbuf,char *fname);

int TXT_Close(WINLIST *w);

void TXT_Draw(WINLIST *w);

void TXT_SetBanner(TEXT_BUF *atbuf,char *aBan);

int TXT_append(TEXT_BUF *aTBuf,char     *aText,int       size);

void TXT_SetCursor(TEXT_BUF_pt atbuf);

void TXT_Redisplay(TEXT_BUF_pt atbuf,SCR_LINES_pt cur_line);

void TXT_Hints(TEXT_BUF_pt atbuf);

void TXT_ReportSize(void);

int TXT_ReallocTview(TEXT_VIEW *atl,long new_size);
int TXT_ReallocTbuf ( TEXT_BUF_pt atbuf, long new_size);

void TXT_SyncWin(TEXT_BUF_pt atbuf);

void TXT_deLink(TEXT_VIEW *atl,int dispose);

void TXT_Link(TEXT_VIEW *atl);

void TXT_DisplayPosition(TEXT_BUF *atbuf);

void TXT_LoadFile(TEXT_BUF *tbuf);

void TXT_RecalcScrLines(TEXT_BUF_pt atbuf,long start_pos);

int TXT_PageFile(TEXT_BUF *tbuf,long offset);

void TXT_UpdateLine(TEXT_BUF_pt atbuf,long pos,int bytes, int size_changed);

void TXT_MakeBlurb(char *title,int  text_length,char *text_pt);

void TXT_Write(TEXT_BUF_pt atbuf,long pos,char *buf,int row, int col, int width);

int TXT_AddAttachment ( TEXT_BUF * tbuf,
                        const int mimeType,
                        const char * mimeSubtype,
                        const char * filename,
                        const char * description,
                        const MIME_ENCODING mimeEncoding );

void TXT_freeAttachments ( TEXT_BUF * tbuf );

#endif /* TEXT_H */
